#ifndef _PLANT_H
#define _PLANT_H

#include <Vec2d.h>
#include <Vec3d.h>
#include <Mat2x2d.h>

using namespace CGLA;

#include <list>

using namespace std;

#include <GL/gl.h>

#include <math.h>

#include "texture.h"
#include "util.h"


// this should be subclassed by the parts of the plant (sub/main stem, flowers, etc)
////
// i should still add some functions like 'gettransform' and getwidth,
// that can be used to provide temportal overloads (wobble/growing) over
// time.. these functions can then be filled in by the derived classes.

// also it makes sense to extract the rotation angle into something
// separate, so we don't have to store the whole transform locally?
Mat2x2d makeRotate( double degrees );

extern Vec2d plantTop;

class PlantSegment
{
protected:
int unique_id;

virtual double getCurrentWidth()
{
	return segmentwidth;
}

virtual double getCurrentLength()
{
	return segmentlength;
}

virtual Mat2x2d getCurrentTransform()
{
	return transform;
	// * makeRotate( sin( age*1.5)*6.0 );
}

protected:
// timestamp at last update ( in seconds )
double now;
// these are the _base_ values for the transform for the current
// segment, and for the segment's _eventual_ width and length. For
// effects like growing and waving, it would be best to override the
// getCurrentWidth get CurrentTransform functions.
Mat2x2d transform;
double segmentlength;
double segmentwidth;
double age;   // age in seconds.
Vec3d color;

// we also need branch counters (number of branches taken, etc)
int branchDepth;
int segmentsSinceRoot;
int segmentsSinceBranch;
list<PlantSegment*> children;

// updated when drawn, the original 'neworigin' value
Vec2d lastOrigin;

// some utility methods for spawning new branches..

public:
bool transformBase;

void setColor( double r,double g,double b )
{
	color = Vec3d( r,g,b );
}

void setColor( Vec3d c )
{
	color = c;
}

void setSegment( float width, float length )
{
	segmentwidth = width;
	segmentlength = length;
}

void setSegmentsSince( int root, int branch )
{
	segmentsSinceRoot = root;
	segmentsSinceBranch = branch;
}

void setBranchDepth( int d )
{
	branchDepth = d;
}

void setTransform( Mat2x2d t )
{
	transform = t;
}

virtual ~PlantSegment()
{
	for ( list<PlantSegment*>::iterator i = children.begin(); i != children.end(); i++ )
		delete *i;
}

PlantSegment()  :
	transformBase( false ),
	segmentsSinceBranch(0),
	segmentsSinceRoot(0),
	age(0.0),
	segmentlength( 1.4 ),
	segmentwidth( 1.0 ),
	transform(1,0,0,1),
	color(0.2,0.7,0.2),
	branchDepth(0)
{
	unique_id = rand() % 1024;
	//      printf("constructed new plant segment\n");
	//transform = makeRotate( 5.0 );
}

virtual void customUpdates( float time, float delta )
{
}

void update( float time, float delta )
{
	now = time;
	// TODO: check whether this reversal of update calls is ok. we need
	// it to pervent crashes fro mhappening (inf recursion) when
	// delta>1.0 or something
	updateChildren(time, delta);
	customUpdates( time, delta );
	age += delta;
}

void updateChildren( float time, float delta )
{
	for ( list<PlantSegment*>::iterator i = children.begin(); i != children.end(); i++ )
	{
		(*i)->update( time, delta );
	}
}

virtual void getTextureCoords( Vec2d & topleft, Vec2d & botright )
{
	topleft = Vec2d( 0,0 );
	botright = Vec2d( 1,1 );
}

virtual void getLocalQuad( Vec2d & topleft, Vec2d & botright )
{
	topleft = Vec2d( -0.5, 1.0 );
	botright = Vec2d( 0.5, 0.0 );
}

virtual void draw( Vec2d origin, Mat2x2d coordsys )
{
	double hwidth = -getCurrentWidth() / 2.0;
	double height = getCurrentLength();
	Mat2x2d ctransform = getCurrentTransform();

	// these coordinates are in our local coordinate system (still has to be pulled through coordsys)
	Vec2d local_topleft, local_botright;
	getLocalQuad( local_topleft, local_botright );
	Vec2d scalevector = Vec2d( getCurrentWidth(), getCurrentLength() );

	Vec2d baseleft = scalevector * Vec2d( local_topleft[0], local_botright[1] );
	Vec2d baseright = scalevector * Vec2d( local_botright[0], local_botright[1] );
	if (transformBase)
	{
		baseleft = ctransform * baseleft;
		baseright = ctransform * baseright;
	}

	Vec2d topleft = ctransform * ( scalevector * Vec2d( local_topleft[0], local_topleft[1] ));
	Vec2d topright = ctransform * ( scalevector * Vec2d( local_botright[0], local_topleft[1] ));

	Vec2d screenBaseLeft = origin+coordsys * baseleft;
	Vec2d screenBaseRight = origin+coordsys * baseright;
	Vec2d screenTopLeft = origin+coordsys * topleft;
	Vec2d screenTopRight = origin+coordsys * topright;

	Vec2d tex_tl, tex_br;
	getTextureCoords( tex_tl, tex_br );

	//      glBegin(GL_QUADS);
	glColor3f(color[0],color[1],color[2]);
	glTexCoord2f( tex_tl[0], tex_tl[1] );
	glVertex2f( screenTopLeft[0], screenTopLeft[1] );
	glTexCoord2f( tex_br[0], tex_tl[1] );
	glVertex2f( screenTopRight[0], screenTopRight[1] );
	glTexCoord2f( tex_br[0], tex_br[1] );
	glVertex2f( screenBaseRight[0], screenBaseRight[1] );
	glTexCoord2f( tex_tl[0], tex_br[1] );
	glVertex2f( screenBaseLeft[0], screenBaseLeft[1] );
	//      glEnd();

	Vec2d neworigin = origin + coordsys* ctransform * Vec2d( 0, height );

	lastOrigin = neworigin;
	for ( list<PlantSegment*>::iterator i = children.begin(); i != children.end(); i++ )
	{
		(*i)->draw( neworigin, coordsys * ctransform );
	}
}

bool pastrange( double current, double checkvalue, double delta )
{
	return (checkvalue >= current && checkvalue < (current+delta));
}
//  double travelledSinceBranch;
};


// specific plants start here:
class GrowingPlantSegment : public PlantSegment
{
public:
virtual double getCurrentWidth()
{
	//      if (age<1.0)
	//        return age/1.0*segmentwidth;

	return segmentwidth;
}

virtual double getCurrentLength()
{
	if (age<1.0)
		return age/1.0*segmentlength;

	return segmentlength;
}

// 'wiegen'
virtual Mat2x2d getCurrentTransform()
{
	double angle=0;
	if ( age*1.5 < 2.0 * M_PI )
		angle = sin( age*1.5)*6.0;

	return transform * makeRotate( angle );
}

};


// simpelplant is a type of tree that has bulbs and flowers
class SimpelPlantFlower : public GrowingPlantSegment
{
public:
virtual double getCurrentWidth()
{
	if (age<1.0)
		return age/1.0*segmentwidth;

	return segmentwidth;
}

virtual Mat2x2d getCurrentTransform()
{
	double angle=0;
	angle = sin( age*1.5)*30.0;

	return transform * makeRotate( angle );
}

SimpelPlantFlower() : GrowingPlantSegment()
{
	segmentlength = 1.25;
	segmentwidth = 1.25;
	//      setColor(1,0,0);
	double r1 = frand();
	setColor( 0.6+ r1*0.6,0.5+r1*0.4,0.2+frand()*0.6 );
	//setColor( 0.3 + 0.8 * frand(),0.2 + 0.4 * frand(),0.3 + 0.3 * frand() );
	//      setColor( frand(), frand(), frand () );
	transformBase = true;
};

virtual void getLocalQuad( Vec2d & topleft, Vec2d & botright )
{
	topleft = Vec2d( -0.5, 0.5 );
	botright = Vec2d( 0.5, -0.5 );
}


virtual void getTextureCoords( Vec2d & topleft, Vec2d & botright )
{
	topleft = Vec2d( 0.25,0 );
	botright = Vec2d( 0.5,0.5 );
}
};

class SimpelPlantBulb : public GrowingPlantSegment
{
public:
SimpelPlantBulb() : GrowingPlantSegment()
{
	segmentlength = 1.25;
	segmentwidth = 1.25;
	setColor(1,0,0);
};


virtual void getTextureCoords( Vec2d & topleft, Vec2d & botright )
{
	topleft = Vec2d( 0.5, 0.1  );
	botright = Vec2d( 0.75, 1.0 );
}
};


class SimpelPlantSegment : public GrowingPlantSegment
{
private:
double randomangle;
double branchgrowtime;
public:

SimpelPlantSegment() : GrowingPlantSegment()
{
	randomangle = 40.0 + 20.0*frand();
	if (rand()%2)
		randomangle *= -1.0;

	branchgrowtime = 1.5 + frand();
};

virtual ~SimpelPlantSegment()
{
};


// 'altijd wiegen'
virtual Mat2x2d getCurrentTransform()
{
	double angle=0;
	//if ( age*1.5 < 2.0 * M_PI )
	angle = sin( age*1.5)*1.0;

	return transform * makeRotate( angle );
}

virtual void getTextureCoords( Vec2d & topleft, Vec2d & botright )
{
	topleft = Vec2d( 0, (segmentsSinceRoot+1) * 0.3  );
	botright = Vec2d( 0.25, (segmentsSinceRoot) * 0.3 );
}

virtual void customUpdates( float time, float delta )
{

	if ( (segmentsSinceRoot+branchDepth*2) < 27 && branchDepth < 4)
	{
		int canbranch = segmentsSinceRoot > 5;

		double anglefac = 1.0  - ( segmentsSinceRoot ) / 30.0;
		if ( pastrange( age, 0.5, delta ) ) // start growing at 0.5 seconds. ONCE
		{
			//              printf("growing a new plantsegment as child\n");
			PlantSegment * ns = new SimpelPlantSegment( );
			ns->setSegment( segmentwidth * 0.95, segmentlength * 0.95);
			ns->setSegmentsSince( segmentsSinceRoot + 1, segmentsSinceBranch + 1 );
			ns->setBranchDepth( branchDepth );
			ns->setColor( 1.05 * color );

			if (segmentsSinceRoot % 3 == 0 && canbranch)
			{
				ns->setTransform( makeRotate( randomangle * anglefac  * ((segmentsSinceRoot/3 % 2) ? -1 : 1) ) );
			}
			else
			{
				ns->setTransform( makeRotate( 40.0 * anglefac  * noise1( unique_id* 100.0 + segmentsSinceRoot*0.3) ));
			}
			/*               ns->segmentlength = segmentlength * 0.95; */
			/*               ns->segmentwidth = segmentwidth * 0.9;  // thinning factor for stems */
			/*               ns->segmentsSinceBranch = segmentsSinceBranch + 1; */
			/*               ns->segmentsSinceRoot = segmentsSinceRoot + 1; */
			// ns->branchDepth = branchDepth;
			children.push_front( ns );
		}

		if ( canbranch && (segmentsSinceRoot % 3 == 0) &&pastrange( age, branchgrowtime, delta ) ) // start growing at 0.5 seconds. ONCE
		{
			PlantSegment * ns = new SimpelPlantSegment( );
			ns->setSegment( segmentwidth * 0.8, segmentlength * 0.8 );
			ns->setColor( 0.8 * color );
			ns->setTransform(  makeRotate( -randomangle * anglefac * ((segmentsSinceRoot/3 % 2) ? -1 : 1) ) );
			ns->setSegmentsSince( segmentsSinceRoot+1, 0 );
			ns->setBranchDepth( branchDepth + 1 );
			ns->transformBase = true;
			children.push_front( ns );
		}
	}
	else
	{
		// end of the branch, lets spawn a flower or a bulb
		if (branchDepth==2 || branchDepth==3) // flower
		{
			if (pastrange( age, 2.5, delta ) )
			{
				SimpelPlantFlower* flower = new SimpelPlantFlower();
				flower->setSegment( 0.6, 0.6 );
				children.push_back( flower );
			}
		}
		else // bulb
		{
			if (pastrange( age, 0.5, delta ) )
			{
				SimpelPlantBulb* bulb = new SimpelPlantBulb();
				bulb->setSegment( segmentwidth * 0.95, segmentlength * 1.85);
				bulb->setSegmentsSince( segmentsSinceRoot + 1, segmentsSinceBranch + 1 );
				bulb->setBranchDepth( branchDepth );
				bulb->setColor( 1.05 * color );
				bulb->setTransform( makeRotate( 10 * ((segmentsSinceRoot/3 % 2) ? -1 : 1) ) );
				children.push_back( bulb );
			}
		}
	}

	// if we have branchdepth 0 and we have no children, then we're the focus of attention!
	if (children.empty() && branchDepth==0)
	{
		if ( !CGLA::isnan(lastOrigin[0]) )
			plantTop = lastOrigin;
	}
}

};

class SimpelPlant : public SimpelPlantSegment
{
private:
Texture tex;

public:
SimpelPlant() : SimpelPlantSegment(), tex("stemflower_debug.png")
{
	setColor( 1,1,1 );
	segmentwidth = 0.3;
	segmentlength = 1.4 / 3.0;
}

void draw( Vec2d origin, Mat2x2d coordsys )
{
	tex.bind();
	glBegin(GL_QUADS);
	SimpelPlantSegment::draw( origin, coordsys );
	glEnd();
	tex.unbind();

}
};


// slingerplant has a main stalk that never stops growing, and has sub-branches and leafs
class SlingerPlantBulb : public GrowingPlantSegment
{
public:
SlingerPlantBulb() : GrowingPlantSegment()
{
	segmentlength = 1.25;
	segmentwidth = 1.25;
	setColor(1,0,0);
};

virtual void getTextureCoords( Vec2d & topleft, Vec2d & botright )
{
	topleft = Vec2d( 0.51, 0.75 );
	botright = Vec2d( 0.74, 0.0 );
	//      topleft = Vec2d( 0.5, 0.1  );
	//      botright = Vec2d( 0.75, 1.0 );
}
};


class SlingerPlantLeaf : public GrowingPlantSegment
{
private:
bool mirror;
int leafType;

public:
virtual double getCurrentWidth()
{
	if (age<1.0)
		return age/1.0*segmentwidth;

	return segmentwidth;
}


SlingerPlantLeaf() : GrowingPlantSegment()
{
	leafType = rand() % 4;
	mirror = rand() % 2;
	segmentlength = 1.25;
	segmentwidth = 1.25;
	setColor(1,0,0);
};

virtual void getTextureCoords( Vec2d & topleft, Vec2d & botright )
{
	if (leafType==0)
	{
		topleft = Vec2d( 0.0, 0.01 );
		botright = Vec2d( 0.125, 0.5 );
	}
	if (leafType==1)
	{
		topleft = Vec2d( 0.125, 0.01 );
		botright = Vec2d( 0.25, 0.5 );
	}
	if (leafType==2)
	{
		topleft = Vec2d( 0.0, 0.51 );
		botright = Vec2d( 0.125, 1.0 );
	}
	if (leafType==3)
	{
		topleft = Vec2d( 0.125, 0.51 );
		botright = Vec2d( 0.25, 1.0 );
	}

	if (mirror)
	{
		double d;

		d = topleft[0];
		topleft[0] = botright[0];
		botright[0] = d;
	}
	//      topleft = Vec2d( 0.5, 0.1  );
	//      botright = Vec2d( 0.75, 1.0 );
}
};


class SlingerPlantFlower : public GrowingPlantSegment
{
private:
int flowerType;
bool mirror;

public:
virtual double getCurrentWidth()
{
	if (age<1.0)
		return age/1.0*segmentwidth;

	return segmentwidth;
}


SlingerPlantFlower() : GrowingPlantSegment()
{
	flowerType = rand() % 2;
	mirror = rand() % 2;
	segmentlength = 1.25;
	segmentwidth = 1.25;
	setColor(1,0,0);
};

virtual void getTextureCoords( Vec2d & topleft, Vec2d & botright )
{
	if (flowerType==0)
	{
		topleft = Vec2d( 0.75, 0.01 );
		botright = Vec2d( 0.75+0.125, 0.5 );
	}
	if (flowerType==1)
	{
		topleft = Vec2d( 0.75, 0.51 );
		botright = Vec2d( 0.75+0.125, 1.0 );
	}

	if (mirror)
	{
		double d;

		d = topleft[0];
		topleft[0] = botright[0];
		botright[0] = d;
	}
	//      topleft = Vec2d( 0.5, 0.1  );
	//      botright = Vec2d( 0.75, 1.0 );
}
};



class SlingerPlantSegment : public GrowingPlantSegment
{
public:
SlingerPlantSegment() : GrowingPlantSegment()
{
};

virtual ~SlingerPlantSegment()
{
};


virtual void getTextureCoords( Vec2d & topleft, Vec2d & botright )
{
	topleft = Vec2d( 0.26, (segmentsSinceRoot+1) * 0.3  );
	botright = Vec2d( 0.49, (segmentsSinceRoot) * 0.3 );
}

virtual void customUpdates( float time, float delta )
{
	//      setColor( 2.0-age, 1.0, 2.0-age );

	if ( branchDepth < 3 && (branchDepth * segmentsSinceBranch < 12))
	{
		double anglefac = 1.0  + 1.0 * branchDepth;
		double growspeed = 1.0;
		if (branchDepth>0)
			growspeed = 2.0;

		if ( pastrange( age, 0.5 / growspeed, delta ) ) // start growing at 0.5 seconds. ONCE
		{
			PlantSegment * ns = new SlingerPlantSegment( );
			if (branchDepth==0)
			{
				ns->setSegment( segmentwidth, segmentlength );
			}
			else
			{
				ns->setSegment( segmentwidth * 0.95, segmentlength * 0.95);
			}

			ns->setSegmentsSince( segmentsSinceRoot + 1, segmentsSinceBranch + 1 );
			ns->setBranchDepth( branchDepth );
			ns->setColor( 1.05 * color );
			if (segmentsSinceRoot % 3 == 0)
			{
				ns->setTransform( makeRotate( 10.0 * anglefac  * ((segmentsSinceRoot/3 % 2) ? -1 : 1) ) );
			}
			else
			{
				ns->setTransform( makeRotate( 50.0 * anglefac  * noise1( segmentsSinceRoot*2.3) ));
			}
			/*               ns->segmentlength = segmentlength * 0.95; */
			/*               ns->segmentwidth = segmentwidth * 0.9;  // thinning factor for stems */
			/*               ns->segmentsSinceBranch = segmentsSinceBranch + 1; */
			/*               ns->segmentsSinceRoot = segmentsSinceRoot + 1; */
			// ns->branchDepth = branchDepth;
			children.push_front( ns );
		}

		if ( (segmentsSinceRoot % 3 == 0) &&pastrange( age, 1.5 / growspeed, delta ) ) // start growing at 0.5 seconds. ONCE
		{
			PlantSegment * ns = new SlingerPlantSegment( );
			ns->setSegment( segmentwidth * 0.8, segmentlength * 0.8 );
			ns->setColor( 0.8 * color );
			ns->setTransform(  makeRotate( -60 * ((segmentsSinceRoot/3 % 2) ? -1 : 1) ) );
			if ( branchDepth == 1 )
			{
				ns->setSegmentsSince( segmentsSinceRoot+1, segmentsSinceBranch+1 );
			}
			else
			{
				ns->setSegmentsSince( segmentsSinceRoot+1, 0 );
			}

			ns->setBranchDepth( branchDepth + 1 );
			ns->transformBase = true;
			children.push_front( ns );
		}
	}
	else
	{
		// end of the branch, lets spawn a flower or a bulb
		//          if (branchDepth==2) // flower
		//            {
/*               if (pastrange( age, 2.5, delta ) ) */
/*                 { */
/*                   SlingerPlantFlower* flower = new SlingerPlantFlower(); */
/*                   flower->setSegment( 0.6, 0.6 ); */
/*                   children.push_back( flower ); */
/*                 } */
//            }
//          else // bulb
		{
			if (pastrange( age, 0.5, delta ) )
			{
				int leafcnt = rand()%3;
				int randomsign = (rand()%2) ? 1 : -1;
				for( int i=0; i<leafcnt; i++)
				{
					SlingerPlantLeaf* leaf = new SlingerPlantLeaf();
					leaf->setSegment( segmentwidth * 4.0, segmentlength * 4.0);
					leaf->setSegmentsSince( segmentsSinceRoot + 1, segmentsSinceBranch + 1 );
					leaf->setBranchDepth( branchDepth );
					leaf->setColor( 1.05 * color );
					leaf->setTransform( makeRotate( 60 * randomsign ) );
					leaf->transformBase = true;
					children.push_back( leaf );
					randomsign *= -1;
				}

				if (leafcnt<2)
				{
					SlingerPlantFlower* flower = new SlingerPlantFlower();
					flower->setSegment( segmentwidth * 4.0, segmentlength * 4.0);
					flower->setSegmentsSince( segmentsSinceRoot + 1, segmentsSinceBranch + 1 );
					flower->setBranchDepth( branchDepth );
					flower->setColor( 0.8 +frand()*0.3,0.3+frand()*0.3,0.6+frand()*0.3 );
					flower->setTransform( makeRotate( 60 * randomsign ) );
					flower->transformBase = true;
					children.push_back( flower );
				}

				SlingerPlantBulb* bulb = new SlingerPlantBulb();
				bulb->setSegment( segmentwidth * 0.95, segmentlength * 1.85);
				bulb->setSegmentsSince( segmentsSinceRoot + 1, segmentsSinceBranch + 1 );
				bulb->setBranchDepth( branchDepth );
				bulb->setColor( 1.05 * color );
				bulb->setTransform( makeRotate( 10 * ((segmentsSinceRoot/3 % 2) ? -1 : 1) ) );
				children.push_back( bulb );

			}
		}
	}

	// snoeien
	/*       if (age > 5.0 && branchDepth>0 && !children.empty()) */
	/*         { */
	/*           delete( *(children.begin()) ); */
	/*           children.clear(); */
	/*         } */

	// if we have branchdepth 0 and we have no children, then we're the focus of attention!
	if (children.empty() && branchDepth==0)
	{
		if ( !CGLA::isnan(lastOrigin[0]) )
			plantTop = lastOrigin;
	}
}

};

class SlingerPlant : public SlingerPlantSegment
{
private:
Texture tex;

public:
SlingerPlant() : SlingerPlantSegment(), tex("slingerplant.png")
{
	setColor( 1,1,1 );
	segmentwidth = 0.3;
	segmentlength = 1.4 / 3.0;
}

void draw( Vec2d origin, Mat2x2d coordsys )
{
	tex.bind();
	glBegin(GL_QUADS);
	SlingerPlantSegment::draw( origin, coordsys );
	glEnd();
	tex.unbind();

}
};



#endif
